package com.fg.scripting.groovy;

import groovy.lang.GroovyClassLoader;
import org.codehaus.groovy.control.CompilerConfiguration;
import org.springframework.aop.Pointcut;
import org.springframework.aop.framework.ProxyFactory;
import org.springframework.aop.interceptor.DebugInterceptor;
import org.springframework.aop.support.DefaultPointcutAdvisor;

import java.util.WeakHashMap;

/**
 * Description
 *
 * @author Jan Novotn, FG Forrest a.s. (c) 2007
 * @version $Id: GroovyClassloaderReloadSupportTest.java,v 1.1 2009/08/10 10:14:48 u_novoj Exp $
 */
public class GroovyClassloaderReloadSupportTest extends AbstractGroovyTest {
	private GroovyClassLoader clsLoader;

	@Override
	protected void setUp() throws Exception {
		super.setUp();
		CompilerConfiguration config = new CompilerConfiguration();		
		config.setMinimumRecompilationInterval(0);
		config.setRecompileGroovySource(true);
		clsLoader = new GroovyClassLoader(
				Thread.currentThread().getContextClassLoader(),
				config
		);
		clsLoader.addURL(rootFSR.getURL());
	}

	public void testReloadLoadGroovyClass() throws Exception {
		//loads class, creates instance of it and calls nonparametrized method helloWorld, all references are
		//then stored in custom class GroovyContext
		GroovyContext ctx = createInstanceAndCallMethod(clsLoader, "com.fg.mock.MainGroovyClass", "helloWorld");
		//initial result of helloWorld method is following
		assertEquals("Hello Universe.", ctx.getGroovyResult());
		//we'll change source file contents so as method returns different value
		modifyGroovyFile(
				"com/fg/mock/MainGroovyClass.groovy",
				"return HELLO_UNIVERSE;",
				"return \"Bye bye\";"
		);
		//existing instances will remain unchanged
		assertEquals("Hello Universe.", ctx.getGroovyMethod().invoke(ctx.getGroovyInstance()));
		//new instances will use new definition
		GroovyContext newCtx = createInstanceAndCallMethod(clsLoader, "com.fg.mock.MainGroovyClass", "helloWorld");
		assertEquals("Bye bye", newCtx.getGroovyResult());
	}

	/**
	 * Run with: -XX:MaxPermSize=1m -Xmx4M
	 * @throws Exception
	 */
	public void testOutOfMemory() throws Exception {
		int iterations = 1000;
		WeakHashMap<Class, Boolean> classes = new WeakHashMap<Class, Boolean>(iterations);
		WeakHashMap<Object, Boolean> instances = new WeakHashMap<Object, Boolean>(iterations);
		WeakHashMap<GroovyClassLoader, Boolean> classloaders = new WeakHashMap<GroovyClassLoader, Boolean>(iterations);
		//uncomment this to end with OOM
		//HashMap<Class, Boolean> classes = new HashMap<Class, Boolean>(iterations);
		//HashMap<Object, Boolean> instances = new HashMap<Object, Boolean>(iterations);

		int i = 0;
		try {
			for (; i < iterations; i++) {
				modifyGroovyFile(
						"com/fg/mock/MainGroovyClass.groovy",
						"return .+;",
						"return \"Hello world " + i + "\";"
				);

				clsLoader = new GroovyClassLoader(Thread.currentThread().getContextClassLoader());
				clsLoader.addURL(rootFSR.getURL());
				classloaders.put(clsLoader, Boolean.TRUE);

				//new instances will use new definition
				GroovyContext ctx = createInstanceAndCallMethod(clsLoader, "com.fg.mock.MainGroovyClass", "helloWorld");
				classes.put(ctx.getGroovyClass(), Boolean.TRUE);
				instances.put(ctx.getGroovyInstance(), Boolean.TRUE);

				if (i % 1000 == 0) {
					System.gc();
					System.out.println(
							(i + 1) + ": retained classes - " + classes.size() +
									", retained instances - " + instances.size() +
									", retained classloaders - " + classloaders.size()
					);
				}
			}
		} catch(OutOfMemoryError e) {
			String msg = "Out of memory with classes - " + classes.size() +
					", retained instances - " + instances.size() +
					", retained classloaders - " + classloaders.size() +
					" on " + (i + 1) + " iterations.";
			System.out.println(msg);
			fail(msg);
		}
	}

	public void testReloadLoadGroovyDependentClass() throws Exception {
		//loads class, creates instance of it and calls nonparametrized method helloWorld, all references are
		//then stored in custom class GroovyContext
		GroovyContext ctx = createInstanceAndCallMethod(clsLoader, "com.fg.mock.MainGroovyClass", "helloWorld");
		//initial result of helloWorld method is following
		assertEquals("Hello Universe.", ctx.getGroovyResult());
		//we'll change source file of the Groovy interface our class implements and use to deduce return value
		//of the helloWorld method call
		modifyGroovyFile(
				"com/fg/mock/GroovyHelloWorldInterface.groovy",
				"HELLO_UNIVERSE = \"Hello Universe.\";",
				"HELLO_UNIVERSE = \"Bye Universe.\";"
		);
		//Groovy class loader doesn't recompile classes if dependent ones change
		GroovyContext newCtx = createInstanceAndCallMethod(clsLoader, "com.fg.mock.MainGroovyClass", "helloWorld");
		assertEquals("Hello Universe.", newCtx.getGroovyResult());
	}

	public void testAopProxy() throws Exception {
		GroovyContext ctx = createInstanceAndCallMethod(clsLoader, "com.fg.mock.MainGroovyClass", "helloWorld");
		//create an AOP proxy for the data provider - proxy implements all provider interfaces
		ProxyFactory factory = new ProxyFactory(ctx.getGroovyInstance());
		factory.addAdvisor(new DefaultPointcutAdvisor(Pointcut.TRUE, new DebugInterceptor()));
		factory.setProxyTargetClass(true);
		factory.setFrozen(true);

		Object proxy = factory.getProxy();
		assertNotNull(proxy);
		assertEquals("Hello Universe.", proxy.getClass().getMethod("helloWorld").invoke(proxy));
	}

}