package com.fg.scripting.groovy;

import groovy.util.GroovyScriptEngine;
import org.springframework.aop.Pointcut;
import org.springframework.aop.framework.ProxyFactory;
import org.springframework.aop.interceptor.DebugInterceptor;
import org.springframework.aop.support.DefaultPointcutAdvisor;

import java.net.URL;
import java.util.WeakHashMap;

/**
 * Description
 *
 * @author Jan Novotn, FG Forrest a.s. (c) 2007
 * @version $Id: GroovyScriptingEngineReloadSupportTest.java,v 1.1 2009/08/10 10:14:48 u_novoj Exp $
 */
public class GroovyScriptingEngineReloadSupportTest extends AbstractGroovyTest {
	protected GroovyScriptEngine engine;

	@Override
	protected void setUp() throws Exception {
		super.setUp();
		engine = new GroovyScriptEngine(
				new URL[] {rootFSR.getURL()}
		);
	}

	public void testReloadGroovyClass() throws Exception {
		//loads class, creates instance of it and calls nonparametrized method helloWorld, all references are
		//then stored in custom class GroovyContext
		GroovyContext ctx = createInstanceAndCallMethod(engine, "com.fg.mock.MainGroovyClass", "helloWorld");
		//initial result of helloWorld method is following
		assertEquals("Hello Universe.", ctx.getGroovyResult());
		//we'll change source file contents so as method returns different value
		modifyGroovyFile(
				"com/fg/mock/MainGroovyClass.groovy",
				"return HELLO_UNIVERSE;",
				"return \"Bye bye\";"
		);
		//existing instances will remain unchanged
		assertEquals("Hello Universe.", ctx.getGroovyMethod().invoke(ctx.getGroovyInstance()));
		//new instances will use new definition
		GroovyContext newCtx = createInstanceAndCallMethod(engine, "com.fg.mock.MainGroovyClass", "helloWorld");
		//and as we see, result gets changed
		assertEquals("Bye bye", newCtx.getGroovyResult());
	}

	/**
	 * Run with: -XX:MaxPermSize=1m -Xmx4M
	 * @throws Exception
	 */
	public void testOutOfMemory() throws Exception {
		int iterations = 1000;
		//uncomment this to have a long and happy life
		WeakHashMap<Class, Boolean> classes = new WeakHashMap<Class, Boolean>(iterations);
		WeakHashMap<Object, Boolean> instances = new WeakHashMap<Object, Boolean>(iterations);
		//this leads to end with OOME
		//HashMap<Class, Boolean> classes = new HashMap<Class, Boolean>(iterations);
		//HashMap<Object, Boolean> instances = new HashMap<Object, Boolean>(iterations);
		//this is test map for querying garbage collection of GroovyScriptEngines
		WeakHashMap<GroovyScriptEngine, Boolean> engines = new WeakHashMap<GroovyScriptEngine, Boolean>(iterations);
		//thousand iterations is enough to fail with 1MB of PermGenSize
		int i = 0;
		try {
			for (; i < iterations; i++) {
				//keep modifying source code file
				modifyGroovyFile(
						"com/fg/mock/MainGroovyClass.groovy",
						"return .+;",
						"return \"Hello world " + i + "\";"
				);
				//each time init new engine (even with the same engine it would early fail)
				engine = new GroovyScriptEngine(
						new URL[] {rootFSR.getURL()}
				);
				engines.put(engine, Boolean.TRUE);
				//new instances will use new definition
				GroovyContext ctx = createInstanceAndCallMethod(engine, "com.fg.mock.MainGroovyClass", "helloWorld");
				//store instances in long living scope
				classes.put(ctx.getGroovyClass(), Boolean.TRUE);
				instances.put(ctx.getGroovyInstance(), Boolean.TRUE);
				//each hundred iterations, run manualy GC (this is not necessary, but makes following line more precise)
				if (i % 100 == 0) {
					System.gc();
					System.out.println(
							(i + 1) + ": retained classes - " + classes.size() +
									", retained instances - " + instances.size() +
									", retained engines - " + engines.size()
					);
				}
			}
		} catch(OutOfMemoryError e) {
			//here we are - with the OOME
			String msg = "Out of memory with classes - " + classes.size() +
					", retained instances - " + instances.size() +
					", retained engines - " + engines.size() +
					" on " + (i + 1) + " iterations.";
			System.out.println(msg);
			fail(msg);
		}
	}

	public void testReloadGroovyDependentClass() throws Exception {
		//loads class, creates instance of it and calls nonparametrized method helloWorld, all references are
		//then stored in custom class GroovyContext
		GroovyContext ctx = createInstanceAndCallMethod(engine, "com.fg.mock.MainGroovyClass", "helloWorld");
		//initial result of helloWorld method is following
		assertEquals("Hello Universe.", ctx.getGroovyResult());
		//we'll change source file of the Groovy interface our class implements and use to deduce return value
		//of the helloWorld method call
		modifyGroovyFile(
				"com/fg/mock/GroovyHelloWorldInterface.groovy",
				"HELLO_UNIVERSE = \"Hello Universe.\";",
				"HELLO_UNIVERSE = \"Bye Universe.\";"
		);
		//even if we didn't change the main class it become reloaded as dependent one was changed
		GroovyContext newCtx = createInstanceAndCallMethod(engine, "com.fg.mock.MainGroovyClass", "helloWorld");
		assertEquals("Bye Universe.", newCtx.getGroovyResult());
	}	

	public void testAopProxy() throws Exception {
		GroovyContext ctx = createInstanceAndCallMethod(engine, "com.fg.mock.MainGroovyClass", "helloWorld");
		//create an AOP proxy for the data provider - proxy implements all provider interfaces
		ProxyFactory factory = new ProxyFactory(ctx.getGroovyInstance());
		factory.addAdvisor(new DefaultPointcutAdvisor(Pointcut.TRUE, new DebugInterceptor()));
		factory.setProxyTargetClass(true);
		factory.setFrozen(true);

		Object proxy = factory.getProxy();
		assertNotNull(proxy);
		assertEquals("Hello Universe.", proxy.getClass().getMethod("helloWorld").invoke(proxy));
	}

}